# Copyright (C) 2002-2006 Stephen Kennedy <stevek@gnome.org>
# Copyright (C) 2010-2019 Kai Willadsen <kai.willadsen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import atexit
import functools
import logging
import os
import shutil
import stat
import sys
import tempfile
from typing import Tuple

from gi.repository import Gdk, Gio, GLib, GObject, Gtk, Pango

from meld import tree
from meld.conf import _
from meld.iohelpers import trash_or_confirm
from meld.melddoc import MeldDoc, open_files_external
from meld.misc import error_dialog, read_pipe_iter
from meld.recent import RecentType
from meld.settings import bind_settings, settings
from meld.ui.vcdialogs import CommitDialog, PushDialog
from meld.vc import _null, get_vcs
from meld.vc._vc import Entry

log = logging.getLogger(__name__)


def cleanup_temp():
    temp_location = tempfile.gettempdir()
    # The strings below will probably end up as debug log, and are deliberately
    # not marked for translation.
    for f in _temp_files:
        try:
            assert (os.path.exists(f) and os.path.isabs(f) and
                    os.path.dirname(f) == temp_location)
            # Windows throws permissions errors if we remove read-only files
            if os.name == "nt":
                os.chmod(f, stat.S_IWRITE)
            os.remove(f)
        except Exception:
            except_str = "{0[0]}: \"{0[1]}\"".format(sys.exc_info())
            print("File \"{0}\" not removed due to".format(f), except_str,
                  file=sys.stderr)
    for f in _temp_dirs:
        try:
            assert (os.path.exists(f) and os.path.isabs(f) and
                    os.path.dirname(f) == temp_location)
            shutil.rmtree(f, ignore_errors=1)
        except Exception:
            except_str = "{0[0]}: \"{0[1]}\"".format(sys.exc_info())
            print("Directory \"{0}\" not removed due to".format(f), except_str,
                  file=sys.stderr)


_temp_dirs, _temp_files = [], []
atexit.register(cleanup_temp)


class ConsoleStream:

    def __init__(self, textview):
        self.textview = textview
        buf = textview.get_buffer()
        self.command_tag = buf.create_tag("command")
        self.command_tag.props.weight = Pango.Weight.BOLD
        self.output_tag = buf.create_tag("output")
        self.error_tag = buf.create_tag("error")
        # FIXME: Need to add this to the gtkrc?
        self.error_tag.props.foreground = "#cc0000"
        self.end_mark = buf.create_mark(None, buf.get_end_iter(),
                                        left_gravity=False)

    def command(self, message):
        self.write(message, self.command_tag)

    def output(self, message):
        self.write(message, self.output_tag)

    def error(self, message):
        self.write(message, self.error_tag)

    def write(self, message, tag):
        if not message:
            return
        buf = self.textview.get_buffer()
        buf.insert_with_tags(buf.get_end_iter(), message, tag)
        self.textview.scroll_mark_onscreen(self.end_mark)


COL_LOCATION, COL_STATUS, COL_OPTIONS, COL_END = \
    list(range(tree.COL_END, tree.COL_END + 4))


class VcTreeStore(tree.DiffTreeStore):
    def __init__(self):
        super().__init__(1, [str] * 5)

    def get_file_path(self, it):
        return self.get_value(it, self.column_index(tree.COL_PATH, 0))


@Gtk.Template(resource_path='/org/gnome/meld/ui/vcview.ui')
class VcView(Gtk.VBox, tree.TreeviewCommon, MeldDoc):

    __gtype_name__ = "VcView"

    __gsettings_bindings__ = (
        ('vc-status-filters', 'status-filters'),
        ('vc-left-is-local', 'left-is-local'),
        ('vc-merge-file-order', 'merge-file-order'),
    )

    close_signal = MeldDoc.close_signal
    create_diff_signal = MeldDoc.create_diff_signal
    file_changed_signal = MeldDoc.file_changed_signal
    label_changed = MeldDoc.label_changed
    move_diff = MeldDoc.move_diff
    tab_state_changed = MeldDoc.tab_state_changed

    status_filters = GObject.Property(
        type=GObject.TYPE_STRV,
        nick="File status filters",
        blurb="Files with these statuses will be shown by the comparison.",
    )
    left_is_local = GObject.Property(type=bool, default=False)
    merge_file_order = GObject.Property(type=str, default="local-merge-remote")

    # Map for inter-tab command() calls
    command_map = {
        'resolve': 'resolve',
    }

    state_actions = {
        'flatten': ('vc-flatten', None),
        'modified': ('vc-status-modified', Entry.is_modified),
        'normal': ('vc-status-normal', Entry.is_normal),
        'unknown': ('vc-status-unknown', Entry.is_nonvc),
        'ignored': ('vc-status-ignored', Entry.is_ignored),
    }

    combobox_vcs = Gtk.Template.Child()
    console_vbox = Gtk.Template.Child()
    consoleview = Gtk.Template.Child()
    emblem_renderer = Gtk.Template.Child()
    extra_column = Gtk.Template.Child()
    extra_renderer = Gtk.Template.Child()
    filelabel = Gtk.Template.Child()
    liststore_vcs = Gtk.Template.Child()
    location_column = Gtk.Template.Child()
    location_renderer = Gtk.Template.Child()
    name_column = Gtk.Template.Child()
    name_renderer = Gtk.Template.Child()
    status_column = Gtk.Template.Child()
    status_renderer = Gtk.Template.Child()
    treeview = Gtk.Template.Child()
    vc_console_vpaned = Gtk.Template.Child()

    def __init__(self):
        super().__init__()
        # FIXME:
        # This unimaginable hack exists because GObject (or GTK+?)
        # doesn't actually correctly chain init calls, even if they're
        # not to GObjects. As a workaround, we *should* just be able to
        # put our class first, but because of Gtk.Template we can't do
        # that if it's a GObject, because GObject doesn't support
        # multiple inheritance and we need to inherit from our Widget
        # parent to make Template work.
        MeldDoc.__init__(self)
        bind_settings(self)

        # Set up per-view action group for top-level menu insertion
        self.view_action_group = Gio.SimpleActionGroup()

        property_actions = (
            ('vc-console-visible', self.console_vbox, 'visible'),
        )
        for action_name, obj, prop_name in property_actions:
            action = Gio.PropertyAction.new(action_name, obj, prop_name)
            self.view_action_group.add_action(action)

        # Manually handle GAction additions
        actions = (
            ('compare', self.action_diff),
            ('find', self.action_find),
            ('next-change', self.action_next_change),
            ('open-external', self.action_open_external),
            ('previous-change', self.action_previous_change),
            ('refresh', self.action_refresh),
            ('vc-add', self.action_add),
            ('vc-commit', self.action_commit),
            ('vc-delete-locally', self.action_delete),
            ('vc-push', self.action_push),
            ('vc-remove', self.action_remove),
            ('vc-resolve', self.action_resolved),
            ('vc-revert', self.action_revert),
            ('vc-update', self.action_update),
        )
        for name, callback in actions:
            action = Gio.SimpleAction.new(name, None)
            action.connect('activate', callback)
            self.view_action_group.add_action(action)

        new_boolean = GLib.Variant.new_boolean
        stateful_actions = (
            ('vc-flatten', self.action_filter_state_change,
                new_boolean('flatten' in self.props.status_filters)),
            ('vc-status-modified', self.action_filter_state_change,
                new_boolean('modified' in self.props.status_filters)),
            ('vc-status-normal', self.action_filter_state_change,
                new_boolean('normal' in self.props.status_filters)),
            ('vc-status-unknown', self.action_filter_state_change,
                new_boolean('unknown' in self.props.status_filters)),
            ('vc-status-ignored', self.action_filter_state_change,
                new_boolean('ignored' in self.props.status_filters)),
        )
        for (name, callback, state) in stateful_actions:
            action = Gio.SimpleAction.new_stateful(name, None, state)
            action.connect('change-state', callback)
            self.view_action_group.add_action(action)

        builder = Gtk.Builder.new_from_resource(
            '/org/gnome/meld/ui/vcview-menus.ui')
        context_menu = builder.get_object('vcview-context-menu')
        self.popup_menu = Gtk.Menu.new_from_model(context_menu)
        self.popup_menu.attach_to_widget(self)

        self.model = VcTreeStore()
        self.connect("style-updated", self.model.on_style_updated)
        self.model.on_style_updated(self)
        self.treeview.set_model(self.model)
        self.treeview.get_selection().connect(
            "changed", self.on_treeview_selection_changed)
        self.treeview.set_search_equal_func(tree.treeview_search_cb, None)
        self.current_path, self.prev_path, self.next_path = None, None, None

        self.name_column.set_attributes(
            self.emblem_renderer,
            icon_name=tree.COL_ICON,
            icon_tint=tree.COL_TINT)
        self.name_column.set_attributes(
            self.name_renderer,
            text=tree.COL_TEXT,
            foreground_rgba=tree.COL_FG,
            style=tree.COL_STYLE,
            weight=tree.COL_WEIGHT,
            strikethrough=tree.COL_STRIKE)
        self.location_column.set_attributes(
            self.location_renderer, markup=COL_LOCATION)
        self.status_column.set_attributes(
            self.status_renderer, markup=COL_STATUS)
        self.extra_column.set_attributes(
            self.extra_renderer, markup=COL_OPTIONS)

        self.consolestream = ConsoleStream(self.consoleview)
        self.location = None
        self.vc = None

        settings.bind('vc-console-visible', self.console_vbox, 'visible',
                      Gio.SettingsBindFlags.DEFAULT)
        settings.bind('vc-console-pane-position', self.vc_console_vpaned,
                      'position', Gio.SettingsBindFlags.DEFAULT)

    def on_container_switch_in_event(self, window):
        super().on_container_switch_in_event(window)
        # FIXME: open-external should be tied to having a treeview selection
        self.set_action_enabled("open-external", True)
        self.scheduler.add_task(self.on_treeview_cursor_changed)

    def on_container_switch_out_event(self, window):
        self.set_action_enabled("open-external", False)
        super().on_container_switch_out_event(window)

    def get_default_vc(self, vcs):
        target_name = self.vc.NAME if self.vc else None

        for i, (name, vc, enabled) in enumerate(vcs):
            if not enabled:
                continue

            if target_name and name == target_name:
                return i

        depths = [len(getattr(vc, 'root', [])) for name, vc, enabled in vcs]
        target_depth = max(depths, default=0)

        for i, (name, vc, enabled) in enumerate(vcs):
            if not enabled:
                continue

            if target_depth and len(vc.root) == target_depth:
                return i

        return 0

    def populate_vcs_for_location(self, location):
        """Display VC plugin(s) that can handle the location"""
        vcs_model = self.combobox_vcs.get_model()
        vcs_model.clear()

        # VC systems can be executed at the directory level, so make sure
        # we're checking for VC support there instead of
        # on a specific file or on deleted/unexisting path inside vc
        location = os.path.abspath(location or ".")
        while not os.path.isdir(location):
            parent_location = os.path.dirname(location)
            if len(parent_location) >= len(location):
                # no existing parent: for example unexisting drive on Windows
                break
            location = parent_location
        else:
            # existing parent directory was found
            for avc, enabled in get_vcs(location):
                err_str = ''
                vc_details = {'name': avc.NAME, 'cmd': avc.CMD}

                if not enabled:
                    # Translators: This error message is shown when no
                    # repository of this type is found.
                    err_str = _("%(name)s (not found)")
                elif not avc.is_installed():
                    # Translators: This error message is shown when a version
                    # control binary isn't installed.
                    err_str = _("%(name)s (%(cmd)s not installed)")
                elif not avc.valid_repo(location):
                    # Translators: This error message is shown when a version
                    # controlled repository is invalid.
                    err_str = _("%(name)s (invalid repository)")

                if err_str:
                    vcs_model.append([err_str % vc_details, avc, False])
                    continue

                vcs_model.append([avc.NAME, avc(location), True])

        default_active = self.get_default_vc(vcs_model)

        if not any(enabled for _, _, enabled in vcs_model):
            # If we didn't get any valid vcs then fallback to null
            null_vcs = _null.Vc(location)
            vcs_model.insert(0, [null_vcs.NAME, null_vcs, True])
            tooltip = _("No valid version control system found in this folder")
        else:
            tooltip = _("Choose which version control system to use")

        self.combobox_vcs.set_tooltip_text(tooltip)
        self.combobox_vcs.set_active(default_active)

    @Gtk.Template.Callback()
    def on_vc_change(self, combobox_vcs):
        active_iter = combobox_vcs.get_active_iter()
        if active_iter is None:
            return
        self.vc = combobox_vcs.get_model()[active_iter][1]
        self._set_location(self.vc.location)

    def set_location(self, location):
        self.populate_vcs_for_location(location)

    def _set_location(self, location):
        self.location = location
        self.current_path = None
        self.model.clear()
        self.filelabel.props.gfile = Gio.File.new_for_path(location)
        it = self.model.add_entries(None, [location])
        self.treeview.grab_focus()
        self.treeview.get_selection().select_iter(it)
        self.model.set_path_state(it, 0, tree.STATE_NORMAL, isdir=1)
        self.recompute_label()
        self.scheduler.remove_all_tasks()

        # If the user is just diffing a file (i.e., not a directory),
        # there's no need to scan the rest of the repository.
        if not os.path.isdir(self.vc.location):
            return

        root = self.model.get_iter_first()
        root_path = self.model.get_path(root)

        try:
            self.model.set_value(
                root, COL_OPTIONS, self.vc.get_commits_to_push_summary())
        except NotImplementedError:
            pass

        self.scheduler.add_task(self.vc.refresh_vc_state)
        self.scheduler.add_task(self._search_recursively_iter(root_path))
        self.scheduler.add_task(self.on_treeview_selection_changed)
        self.scheduler.add_task(self.on_treeview_cursor_changed)

    def get_comparison(self):
        uris = [Gio.File.new_for_path(self.location)]
        return RecentType.VersionControl, uris

    def recompute_label(self):
        self.label_text = os.path.basename(self.location)
        self.tooltip_text = "\n".join((
            # TRANSLATORS: This is the name of the version control
            # system being used, e.g., "Git" or "Subversion"
            _("{vc} comparison:").format(vc=self.vc.NAME),
            self.location,
        ))
        self.label_changed.emit(self.label_text, self.tooltip_text)

    def _search_recursively_iter(self, start_path, replace=False):

        # Initial yield so when we add this to our tasks, we don't
        # create iterators that may be invalidated.
        yield _("Scanning repository")

        if replace:
            # Replace the row at start_path with a new, empty row ready
            # to be filled.
            old_iter = self.model.get_iter(start_path)
            file_path = self.model.get_file_path(old_iter)
            new_iter = self.model.insert_after(None, old_iter)
            self.model.set_value(new_iter, tree.COL_PATH, file_path)
            self.model.set_path_state(new_iter, 0, tree.STATE_NORMAL, True)
            self.model.remove(old_iter)

        iterstart = self.model.get_iter(start_path)
        rootname = self.model.get_file_path(iterstart)
        display_prefix = len(rootname) + 1
        symlinks_followed = set()
        todo = [(self.model.get_path(iterstart), rootname)]

        flattened = 'flatten' in self.props.status_filters
        active_actions = [
            self.state_actions.get(k) for k in self.props.status_filters]
        filters = [a[1] for a in active_actions if a and a[1]]

        while todo:
            # This needs to happen sorted and depth-first in order for our row
            # references to remain valid while we traverse.
            todo.sort()
            treepath, path = todo.pop(0)
            it = self.model.get_iter(treepath)
            yield _("Scanning %s") % path[display_prefix:]

            entries = self.vc.get_entries(path)
            entries = [e for e in entries if any(f(e) for f in filters)]
            entries = sorted(entries, key=lambda e: e.name)
            entries = sorted(entries, key=lambda e: not e.isdir)
            for e in entries:
                if e.isdir and e.is_present():
                    try:
                        st = os.lstat(e.path)
                    # Covers certain unreadable symlink cases; see bgo#585895
                    except OSError as err:
                        error_string = "%r: %s" % (e.path, err.strerror)
                        self.model.add_error(it, error_string, 0)
                        continue

                    if stat.S_ISLNK(st.st_mode):
                        key = (st.st_dev, st.st_ino)
                        if key in symlinks_followed:
                            continue
                        symlinks_followed.add(key)

                    if flattened:
                        if e.state != tree.STATE_IGNORED:
                            # If directory state is changed, render it in
                            # in flattened mode.
                            if e.state != tree.STATE_NORMAL:
                                child = self.model.add_entries(it, [e.path])
                                self._update_item_state(child, e)
                            todo.append((Gtk.TreePath.new_first(), e.path))
                        continue

                child = self.model.add_entries(it, [e.path])
                if e.isdir and e.state != tree.STATE_IGNORED:
                    todo.append((self.model.get_path(child), e.path))
                self._update_item_state(child, e)

            if not flattened:
                if not entries:
                    self.model.add_empty(it, _("(Empty)"))
                elif any(e.state != tree.STATE_NORMAL for e in entries):
                    self.treeview.expand_to_path(treepath)

        self.treeview.expand_row(Gtk.TreePath.new_first(), False)
        self.treeview.set_cursor(Gtk.TreePath.new_first())

    # TODO: This doesn't fire when the user selects a shortcut folder
    @Gtk.Template.Callback()
    def on_file_selected(
            self, button: Gtk.Button, pane: int, file: Gio.File) -> None:

        path = file.get_path()
        self.set_location(path)

    def on_delete_event(self):
        self.scheduler.remove_all_tasks()
        self.close_signal.emit(0)
        return Gtk.ResponseType.OK

    @Gtk.Template.Callback()
    def on_row_activated(self, treeview, path, tvc):
        it = self.model.get_iter(path)
        if self.model.iter_has_child(it):
            if self.treeview.row_expanded(path):
                self.treeview.collapse_row(path)
            else:
                self.treeview.expand_row(path, False)
        else:
            path = self.model.get_file_path(it)
            if not self.model.is_folder(it, 0, path):
                self.run_diff(path)

    def run_diff(self, path):
        if os.path.isdir(path):
            self.create_diff_signal.emit([Gio.File.new_for_path(path)], {})
            return

        basename = os.path.basename(path)
        meta = {
            'parent': self,
            'prompt_resolve': False,
        }

        # May have removed directories in list.
        vc_entry = self.vc.get_entry(path)
        if vc_entry and vc_entry.state == tree.STATE_CONFLICT and \
                hasattr(self.vc, 'get_path_for_conflict'):
            local_label = _("%s — local") % basename
            remote_label = _("%s — remote") % basename

            # We create new temp files for other, base and this, and
            # then set the output to the current file.
            if self.props.merge_file_order == "local-merge-remote":
                conflicts = (tree.CONFLICT_THIS, tree.CONFLICT_MERGED,
                             tree.CONFLICT_OTHER)
                meta['labels'] = (local_label, None, remote_label)
                meta['tablabel'] = _("%s (local, merge, remote)") % basename
            else:
                conflicts = (tree.CONFLICT_OTHER, tree.CONFLICT_MERGED,
                             tree.CONFLICT_THIS)
                meta['labels'] = (remote_label, None, local_label)
                meta['tablabel'] = _("%s (remote, merge, local)") % basename
            diffs = [self.vc.get_path_for_conflict(path, conflict=c)
                     for c in conflicts]
            temps = [p for p, is_temp in diffs if is_temp]
            diffs = [p for p, is_temp in diffs]
            kwargs = {
                'auto_merge': False,
                'merge_output': Gio.File.new_for_path(path),
            }
            meta['prompt_resolve'] = True
        else:
            remote_label = _("%s — repository") % basename
            comp_path = self.vc.get_path_for_repo_file(path)
            temps = [comp_path]
            if self.props.left_is_local:
                diffs = [path, comp_path]
                meta['labels'] = (None, remote_label)
                meta['tablabel'] = _("%s (working, repository)") % basename
            else:
                diffs = [comp_path, path]
                meta['labels'] = (remote_label, None)
                meta['tablabel'] = _("%s (repository, working)") % basename
            kwargs = {}
        kwargs['meta'] = meta

        for temp_file in temps:
            os.chmod(temp_file, 0o444)
            _temp_files.append(temp_file)

        self.create_diff_signal.emit(
            [Gio.File.new_for_path(d) for d in diffs],
            kwargs,
        )

    def get_filter_visibility(self) -> Tuple[bool, bool, bool]:
        return False, False, True

    def action_filter_state_change(self, action, value):
        action.set_state(value)

        active_filters = [
            k for k, (action_name, fn) in self.state_actions.items()
            if self.get_action_state(action_name)
        ]

        if set(active_filters) == set(self.props.status_filters):
            return

        self.props.status_filters = active_filters
        self.refresh()

    def on_treeview_selection_changed(self, selection=None):
        if selection is None:
            selection = self.treeview.get_selection()
        model, rows = selection.get_selected_rows()
        paths = [self.model.get_file_path(model.get_iter(r)) for r in rows]
        states = [self.model.get_state(model.get_iter(r), 0) for r in rows]
        path_states = dict(zip(paths, states))

        valid_actions = self.vc.get_valid_actions(path_states)
        action_sensitivity = {
            'compare': 'compare' in valid_actions,
            'vc-add': 'add' in valid_actions,
            'vc-commit': 'commit' in valid_actions,
            'vc-delete-locally': bool(paths) and self.vc.root not in paths,
            'vc-push': 'push' in valid_actions,
            'vc-remove': 'remove' in valid_actions,
            'vc-resolve': 'resolve' in valid_actions,
            'vc-revert': 'revert' in valid_actions,
            'vc-update': 'update' in valid_actions,
        }
        for action, sensitivity in action_sensitivity.items():
            self.set_action_enabled(action, sensitivity)

    def _get_selected_files(self):
        model, rows = self.treeview.get_selection().get_selected_rows()
        sel = [self.model.get_file_path(self.model.get_iter(r)) for r in rows]
        # Remove empty entries and trailing slashes
        return [x[-1] != "/" and x or x[:-1] for x in sel if x is not None]

    def _command_iter(self, command, files, refresh, working_dir):
        """An iterable that runs a VC command on a set of files

        This method is intended to be used as a scheduled task, with
        standard out and error output displayed in this view's
        consolestream.
        """

        def shelljoin(command):
            def quote(s):
                return '"%s"' % s if len(s.split()) > 1 else s
            return " ".join(quote(tok) for tok in command)

        files = [os.path.relpath(f, working_dir) for f in files]
        msg = shelljoin(command + files) + " (in %s)\n" % working_dir
        self.consolestream.command(msg)
        readiter = read_pipe_iter(
            command + files, workdir=working_dir,
            errorstream=self.consolestream)
        try:
            result = next(readiter)
            while not result:
                yield 1
                result = next(readiter)
        except IOError as err:
            error_dialog(
                "Error running command",
                "While running '%s'\nError: %s" % (msg, err))
            result = (1, "")

        returncode, output = result
        self.consolestream.output(output + "\n")

        if returncode:
            self.console_vbox.show()

        if refresh:
            refresh = functools.partial(self.refresh_partial, working_dir)
            GLib.idle_add(refresh)

    def has_command(self, command):
        vc_command = self.command_map.get(command)
        return vc_command and hasattr(self.vc, vc_command)

    def command(self, command, files, sync=False):
        """
        Run a command against this view's version control subsystem

        This is the intended way for things outside of the VCView to
        call in to version control methods, e.g., to mark a conflict as
        resolved from a file comparison.

        :param command: The version control command to run, taken from
            keys in `VCView.command_map`.
        :param files: File parameters to the command as paths
        :param sync: If True, the command will be executed immediately
            (as opposed to being run by the idle scheduler).
        """
        if not self.has_command(command):
            log.error("Couldn't understand command %s", command)
            return

        if not isinstance(files, list):
            log.error("Invalid files argument to '%s': %r", command, files)
            return

        runner = self.runner if not sync else self.sync_runner
        command = getattr(self.vc, self.command_map[command])
        command(runner, files)

    def runner(self, command, files, refresh, working_dir):
        """Schedule a version control command to run as an idle task"""
        self.scheduler.add_task(
            self._command_iter(command, files, refresh, working_dir))

    def sync_runner(self, command, files, refresh, working_dir):
        """Run a version control command immediately"""
        for it in self._command_iter(command, files, refresh, working_dir):
            pass

    def action_update(self, *args):
        self.vc.update(self.runner)

    def action_push(self, *args):
        response = PushDialog(self).run()
        if response == Gtk.ResponseType.OK:
            self.vc.push(self.runner)

    def action_commit(self, *args):
        response, commit_msg = CommitDialog(self).run()
        if response == Gtk.ResponseType.OK:
            self.vc.commit(
                self.runner, self._get_selected_files(), commit_msg)

    def action_add(self, *args):
        self.vc.add(self.runner, self._get_selected_files())

    def action_remove(self, *args):
        selected = self._get_selected_files()
        if any(os.path.isdir(p) for p in selected):
            # TODO: Improve and reuse this dialog for the non-VC delete action
            dialog = Gtk.MessageDialog(
                parent=self.get_toplevel(),
                flags=(Gtk.DialogFlags.MODAL |
                       Gtk.DialogFlags.DESTROY_WITH_PARENT),
                type=Gtk.MessageType.WARNING,
                message_format=_("Remove folder and all its files?"))
            dialog.format_secondary_text(
                _("This will remove all selected files and folders, and all "
                  "files within any selected folders, from version control."))

            dialog.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
            dialog.add_button(_("_Remove"), Gtk.ResponseType.OK)
            response = dialog.run()
            dialog.destroy()
            if response != Gtk.ResponseType.OK:
                return

        self.vc.remove(self.runner, selected)

    def action_resolved(self, *args):
        self.vc.resolve(self.runner, self._get_selected_files())

    def action_revert(self, *args):
        self.vc.revert(self.runner, self._get_selected_files())

    def action_delete(self, *args):
        files = self._get_selected_files()
        for name in files:
            gfile = Gio.File.new_for_path(name)

            try:
                trash_or_confirm(gfile)
            except Exception as e:
                error_dialog(
                    _("Error deleting {}").format(
                        GLib.markup_escape_text(gfile.get_parse_name()),
                    ),
                    str(e),
                )

        workdir = os.path.dirname(os.path.commonprefix(files))
        self.refresh_partial(workdir)

    def action_diff(self, *args):
        # TODO: Review the compare/diff action. It doesn't really add much
        # over activate, since the folder compare doesn't work and hasn't
        # for... a long time.
        files = self._get_selected_files()
        for f in files:
            self.run_diff(f)

    def action_open_external(self, *args):
        open_files_external(self._get_selected_files())

    def refresh(self):
        root = self.model.get_iter_first()
        if root is None:
            return
        self.set_location(self.model.get_file_path(root))

    def refresh_partial(self, where):
        if not self.get_action_state('vc-flatten'):
            it = self.find_iter_by_name(where)
            if not it:
                return
            path = self.model.get_path(it)

            self.treeview.grab_focus()
            self.vc.refresh_vc_state(where)
            self.scheduler.add_task(
                self._search_recursively_iter(path, replace=True))
            self.scheduler.add_task(self.on_treeview_selection_changed)
            self.scheduler.add_task(self.on_treeview_cursor_changed)
        else:
            # XXX fixme
            self.refresh()

    def _update_item_state(self, it, entry):
        self.model.set_path_state(it, 0, entry.state, entry.isdir)

        location = Gio.File.new_for_path(self.vc.location)
        parent = Gio.File.new_for_path(entry.path).get_parent()
        display_location = location.get_relative_path(parent)

        self.model.set_value(it, COL_LOCATION, display_location)
        self.model.set_value(it, COL_STATUS, entry.get_status())
        self.model.set_value(it, COL_OPTIONS, entry.options)

    def on_file_changed(self, filename):
        it = self.find_iter_by_name(filename)
        if it:
            path = self.model.get_file_path(it)
            self.vc.refresh_vc_state(path)
            entry = self.vc.get_entry(path)
            self._update_item_state(it, entry)

    def find_iter_by_name(self, name):
        it = self.model.get_iter_first()
        path = self.model.get_file_path(it)
        while it:
            if name == path:
                return it
            elif name.startswith(path):
                child = self.model.iter_children(it)
                while child:
                    path = self.model.get_file_path(child)
                    if name == path:
                        return child
                    elif name.startswith(path):
                        break
                    else:
                        child = self.model.iter_next(child)
                it = child
            else:
                break
        return None

    @Gtk.Template.Callback()
    def on_consoleview_populate_popup(self, textview, menu):
        buf = textview.get_buffer()
        clear_action = Gtk.MenuItem.new_with_label(_("Clear"))
        clear_action.connect(
            "activate", lambda *args: buf.delete(*buf.get_bounds()))
        menu.insert(clear_action, 0)
        menu.insert(Gtk.SeparatorMenuItem(), 1)
        menu.show_all()

    @Gtk.Template.Callback()
    def on_treeview_popup_menu(self, treeview):
        return tree.TreeviewCommon.on_treeview_popup_menu(self, treeview)

    @Gtk.Template.Callback()
    def on_treeview_button_press_event(self, treeview, event):
        return tree.TreeviewCommon.on_treeview_button_press_event(
            self, treeview, event)

    @Gtk.Template.Callback()
    def on_treeview_cursor_changed(self, *args):
        cursor_path, cursor_col = self.treeview.get_cursor()
        if not cursor_path:
            self.set_action_enabled("previous-change", False)
            self.set_action_enabled("next-change", False)
            self.current_path = cursor_path
            return

        # If invoked directly rather than through a callback, we always check
        if not args:
            skip = False
        else:
            try:
                old_cursor = self.model.get_iter(self.current_path)
            except (ValueError, TypeError):
                # An invalid path gives ValueError; None gives a TypeError
                skip = False
            else:
                # We can skip recalculation if the new cursor is between
                # the previous/next bounds, and we weren't on a changed row
                state = self.model.get_state(old_cursor, 0)
                if state not in (tree.STATE_NORMAL, tree.STATE_EMPTY):
                    skip = False
                else:
                    if self.prev_path is None and self.next_path is None:
                        skip = True
                    elif self.prev_path is None:
                        skip = cursor_path < self.next_path
                    elif self.next_path is None:
                        skip = self.prev_path < cursor_path
                    else:
                        skip = self.prev_path < cursor_path < self.next_path

        if not skip:
            prev, next_ = self.model._find_next_prev_diff(cursor_path)
            self.prev_path, self.next_path = prev, next_
            self.set_action_enabled("previous-change", prev is not None)
            self.set_action_enabled("next-change", next_ is not None)
        self.current_path = cursor_path

    def next_diff(self, direction):
        if direction == Gdk.ScrollDirection.UP:
            path = self.prev_path
        else:
            path = self.next_path
        if path:
            self.treeview.expand_to_path(path)
            self.treeview.set_cursor(path)
        else:
            self.error_bell()

    def action_previous_change(self, *args):
        self.next_diff(Gdk.ScrollDirection.UP)

    def action_next_change(self, *args):
        self.next_diff(Gdk.ScrollDirection.DOWN)

    def action_refresh(self, *args):
        self.set_location(self.location)

    def action_find(self, *args):
        self.treeview.emit("start-interactive-search")

    def auto_compare(self):
        modified_states = (tree.STATE_MODIFIED, tree.STATE_CONFLICT)
        for it in self.model.state_rows(modified_states):
            row_paths = self.model.value_paths(it)
            paths = [p for p in row_paths if os.path.exists(p)]
            self.run_diff(paths[0])


VcView.set_css_name('meld-vc-view')




from unittest import mock

import pytest
from gi.repository import GtkSource


@pytest.mark.parametrize("text, newline, expected_text", [
    # For the following tests, newlines and text match
    # Basic CRLF tests
    ("ree\r\neee", GtkSource.NewlineType.CR_LF, 'ree'),
    ("ree\r\neee\r\n", GtkSource.NewlineType.CR_LF, 'ree\r\neee'),
    # Basic CR tests
    ("ree\neee", GtkSource.NewlineType.CR, 'ree'),
    ("ree\neee\n", GtkSource.NewlineType.CR, 'ree\neee'),
    # Basic LF tests
    ("ree\reee", GtkSource.NewlineType.LF, 'ree'),
    ("ree\reee\r", GtkSource.NewlineType.LF, 'ree\reee'),

    # Mismatched newline and text
    ("ree\r\neee", GtkSource.NewlineType.CR, 'ree'),

    # Mismatched newline types within text
    ("ree\r\neee\n", GtkSource.NewlineType.CR_LF, 'ree\r\neee'),
    ("ree\r\neee\nqqq", GtkSource.NewlineType.CR_LF, 'ree\r\neee'),
    ("ree\r\neee\nqqq\r\n", GtkSource.NewlineType.CR_LF, 'ree\r\neee\nqqq'),
])
def test_delete_last_line_crlf(text, newline, expected_text):
    import meld.meldbuffer
    from meld.filediff import FileDiff
    from meld.matchers.myers import DiffChunk

    filediff = mock.Mock(FileDiff)

    with mock.patch('meld.meldbuffer.bind_settings', mock.DEFAULT):
        meldbuffer = meld.meldbuffer.MeldBuffer()
        meldbuffer.set_text(text)

    def make_last_line_chunk(buf):
        end = buf.get_line_count()
        last = end - 1
        return DiffChunk('delete', last, end, last, end)

    start, end = meldbuffer.get_bounds()
    buf_text = meldbuffer.get_text(start, end, False)
    print(repr(buf_text))

    with mock.patch.object(
            meldbuffer.data.sourcefile,
            'get_newline_type', return_value=newline):
        filediff.textbuffer = [meldbuffer]
        filediff.textview = [mock.Mock()]
        FileDiff.delete_chunk(filediff, 0, make_last_line_chunk(meldbuffer))

    start, end = meldbuffer.get_bounds()
    buf_text = meldbuffer.get_text(start, end, False)
    print(repr(buf_text))
    assert buf_text == expected_text



from gi.repository import Gtk

from meld.accelerators import VIEW_ACCELERATORS


def test_accelerator_parse():
    for accel_strings in VIEW_ACCELERATORS.values():
        if isinstance(accel_strings, str):
            accel_strings = [accel_strings]

        for accel_string in accel_strings:
            key, mods = Gtk.accelerator_parse(accel_string)
            assert key


def test_accelerator_duplication():
    accels = set()

    allowed_duplicates = {
        # Allowed because they're different copy actions across views
        Gtk.accelerator_parse("<Alt>Left"),
        Gtk.accelerator_parse("<Alt>Right"),
        # Allowed because they're different panel show/hide across views
        Gtk.accelerator_parse("F9"),
    }

    for accel_strings in VIEW_ACCELERATORS.values():
        if isinstance(accel_strings, str):
            accel_strings = [accel_strings]

        for accel_string in accel_strings:
            accel = Gtk.accelerator_parse(accel_string)

            if accel not in allowed_duplicates:
                assert (
                    accel not in accels
                ), f"Duplicate accelerator {accel_string}"
            accels.add(accel)



import pytest


@pytest.mark.parametrize('txt, expected', [
    # blank to be equal blank
    (b'', b''),
    # one line with spaces
    (b' ', b' '),
    # two lines empty
    (b'\n', b''),
    (b'\n ', b' '),
    (b' \n', b' '),
    (b' \n ', b' \n '),
    # tree lines empty
    (b'\n\n', b''),
    (b'\n\n ', b' '),
    (b'\n \n', b' '),
    (b'\n \n ', b' \n '),
    (b' \n \n ', b' \n \n '),
    # one line with space and content
    (b' content', b' content'),
    # empty line between content
    (b'content\n\ncontent', b'content\ncontent'),
    # multiple leading and trailing newlines
    (b'\n\n\ncontent\n\n\ncontent\n\n\n', b'content\ncontent'),
])
def test_remove_blank_lines(txt, expected):
    from meld.dirdiff import remove_blank_lines

    result = remove_blank_lines(txt)
    assert result == expected




from typing import Optional

from gi.repository import Gio, GObject, Gtk


class MeldFileButton(Gtk.Button):
    __gtype_name__ = "MeldFileButton"

    file: Optional[Gio.File] = GObject.Property(
        type=Gio.File,
        nick="Most recently selected file",
    )

    pane: int = GObject.Property(
        type=int,
        nick="Index of pane associated with this file selector",
        flags=(
            GObject.ParamFlags.READWRITE |
            GObject.ParamFlags.CONSTRUCT_ONLY
        ),
    )

    action: Gtk.FileChooserAction = GObject.Property(
        type=Gtk.FileChooserAction,
        nick="File selector action",
        flags=(
            GObject.ParamFlags.READWRITE |
            GObject.ParamFlags.CONSTRUCT_ONLY
        ),
        default=Gtk.FileChooserAction.OPEN,
    )

    local_only: bool = GObject.Property(
        type=bool,
        nick="Whether selected files should be limited to local file:// URIs",
        flags=(
            GObject.ParamFlags.READWRITE |
            GObject.ParamFlags.CONSTRUCT_ONLY
        ),
        default=True,
    )

    dialog_label: str = GObject.Property(
        type=str,
        nick="Label for the file selector dialog",
        flags=(
            GObject.ParamFlags.READWRITE |
            GObject.ParamFlags.CONSTRUCT_ONLY
        ),
    )

    @GObject.Signal('file-selected')
    def file_selected_signal(self, pane: int, file: Gio.File) -> None:
        ...

    icon_action_map = {
        Gtk.FileChooserAction.OPEN: "document-open-symbolic",
        Gtk.FileChooserAction.SELECT_FOLDER: "folder-open-symbolic",
    }

    def do_realize(self) -> None:
        Gtk.Button.do_realize(self)

        image = Gtk.Image.new_from_icon_name(
            self.icon_action_map[self.action], Gtk.IconSize.BUTTON)
        self.set_image(image)

    def do_clicked(self) -> None:
        dialog = Gtk.FileChooserNative(
            title=self.dialog_label,
            transient_for=self.get_toplevel(),
            action=self.action,
            local_only=self.local_only
        )

        if self.file and self.file.get_path():
            dialog.set_file(self.file)

        response = dialog.run()
        gfile = dialog.get_file()
        dialog.destroy()

        if response != Gtk.ResponseType.ACCEPT or not gfile:
            return

        self.file = gfile
        self.file_selected_signal.emit(self.pane, self.file)



# Copyright (C) 2002-2009 Stephen Kennedy <stevek@gnome.org>
# Copyright (C) 2012-2014 Kai Willadsen <kai.willadsen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import ClassVar, Optional

from gi.repository import GObject, Gtk, GtkSource


@Gtk.Template(resource_path='/org/gnome/meld/ui/findbar.ui')
class FindBar(Gtk.Grid):

    __gtype_name__ = 'FindBar'

    find_entry = Gtk.Template.Child()
    find_next_button = Gtk.Template.Child()
    find_previous_button = Gtk.Template.Child()
    match_case = Gtk.Template.Child()
    regex = Gtk.Template.Child()
    replace_all_button = Gtk.Template.Child()
    replace_button = Gtk.Template.Child()
    replace_entry = Gtk.Template.Child()
    whole_word = Gtk.Template.Child()
    wrap_box = Gtk.Template.Child()

    replace_mode = GObject.Property(type=bool, default=False)
    _cached_search: ClassVar[Optional[str]] = None

    @GObject.Signal(
        name='activate-secondary',
        flags=(
            GObject.SignalFlags.RUN_FIRST |
            GObject.SignalFlags.ACTION
        ),
    )
    def activate_secondary(self) -> None:
        self._find_text(backwards=True)

    def __init__(self, parent):
        super().__init__()

        self.search_context = None
        self.notify_id = None
        self.set_text_view(None)

        # Create and bind our GtkSourceSearchSettings
        settings = GtkSource.SearchSettings()
        self.match_case.bind_property('active', settings, 'case-sensitive')
        self.whole_word.bind_property('active', settings, 'at-word-boundaries')
        self.regex.bind_property('active', settings, 'regex-enabled')
        self.find_entry.bind_property('text', settings, 'search-text')
        settings.set_wrap_around(True)
        self.search_settings = settings

        # Bind visibility and layout for find-and-replace mode
        self.bind_property('replace_mode', self.replace_entry, 'visible')
        self.bind_property('replace_mode', self.replace_all_button, 'visible')
        self.bind_property('replace_mode', self.replace_button, 'visible')
        self.bind_property(
            'replace_mode', self, 'row-spacing', GObject.BindingFlags.DEFAULT,
            lambda binding, replace_mode: 6 if replace_mode else 0)

    def hide(self):
        self.set_text_view(None)
        self.wrap_box.set_visible(False)
        Gtk.Widget.hide(self)

    def update_match_state(self, *args):
        # Note that -1 here implies that the search is still running
        no_matches = (
            self.search_context.props.occurrences_count == 0 and
            self.search_settings.props.search_text
        )
        style_context = self.find_entry.get_style_context()
        if no_matches:
            style_context.add_class(Gtk.STYLE_CLASS_ERROR)
        else:
            style_context.remove_class(Gtk.STYLE_CLASS_ERROR)

    def set_text_view(self, textview):
        self.textview = textview
        if textview is not None:
            self.search_context = GtkSource.SearchContext.new(
                textview.get_buffer(), self.search_settings)
            self.search_context.set_highlight(True)
            self.notify_id = self.search_context.connect(
                'notify::occurrences-count', self.update_match_state)
        else:
            if self.notify_id:
                self.search_context.disconnect(self.notify_id)
                self.notify_id = None
            self.search_context = None

    def start_find(self, *, textview: Gtk.TextView, replace: bool, text: str):
        self.replace_mode = replace
        self.set_text_view(textview)
        if text:
            self.find_entry.set_text(text)
            FindBar._cached_search = text
        elif FindBar._cached_search:
            self.find_entry.set_text(FindBar._cached_search)
        self.show()
        self.find_entry.grab_focus()

    def start_find_next(self, textview):
        self.set_text_view(textview)
        self._find_text()

    def start_find_previous(self, textview):
        self.set_text_view(textview)
        self._find_text(backwards=True)

    @Gtk.Template.Callback()
    def on_find_next_button_clicked(self, button):
        self._find_text()

    @Gtk.Template.Callback()
    def on_find_previous_button_clicked(self, button):
        self._find_text(backwards=True)

    @Gtk.Template.Callback()
    def on_replace_button_clicked(self, entry):
        buf = self.textview.get_buffer()
        oldsel = buf.get_selection_bounds()
        match = self._find_text(0)
        newsel = buf.get_selection_bounds()
        # Only replace if there is an already-selected match at the cursor
        if (match and oldsel and oldsel[0].equal(newsel[0]) and
                oldsel[1].equal(newsel[1])):
            self.search_context.replace(
                newsel[0], newsel[1], self.replace_entry.get_text(), -1)
            self._find_text(0)

    @Gtk.Template.Callback()
    def on_replace_all_button_clicked(self, entry):
        buf = self.textview.get_buffer()
        saved_insert = buf.create_mark(
            None, buf.get_iter_at_mark(buf.get_insert()), True)
        self.search_context.replace_all(self.replace_entry.get_text(), -1)
        if not saved_insert.get_deleted():
            buf.place_cursor(buf.get_iter_at_mark(saved_insert))
            self.textview.scroll_to_mark(
                buf.get_insert(), 0.25, True, 0.5, 0.5)

    @Gtk.Template.Callback()
    def on_toggle_replace_button_clicked(self, button):
        self.replace_mode = not self.replace_mode

    @Gtk.Template.Callback()
    def on_find_entry_changed(self, entry):
        FindBar._cached_search = entry.get_text()
        self._find_text(0)

    @Gtk.Template.Callback()
    def on_stop_search(self, search_entry):
        self.hide()

    def _find_text(self, start_offset=1, backwards=False):
        if not self.textview or not self.search_context:
            return

        buf = self.textview.get_buffer()
        insert = buf.get_iter_at_mark(buf.get_insert())

        start, end = buf.get_bounds()
        self.wrap_box.set_visible(False)
        if not backwards:
            insert.forward_chars(start_offset)
            match, start, end, wrapped = self.search_context.forward(insert)
        else:
            match, start, end, wrapped = self.search_context.backward(insert)

        if match:
            self.wrap_box.set_visible(wrapped)
            buf.place_cursor(start)
            buf.move_mark(buf.get_selection_bound(), end)
            self.textview.scroll_to_mark(
                buf.get_insert(), 0.25, True, 0.5, 0.5)
            return True
        else:
            buf.place_cursor(buf.get_iter_at_mark(buf.get_insert()))
            self.wrap_box.set_visible(False)


FindBar.set_css_name('meld-find-bar')




from unittest import mock

import pytest
from gi.repository import Gtk


@pytest.mark.parametrize("text, ignored_ranges, expected_text", [
    #    0123456789012345678901234567890123456789012345678901234567890123456789
    # Matching without groups
    (
        "# asdasdasdasdsad",
        [(0, 17)],
        "",
    ),
    # Matching with single group
    (
        "asdasdasdasdsab",
        [(1, 14)],
        "ab",
    ),
    # Matching with multiple groups
    (
        "xasdyasdz",
        [(1, 4), (5, 8)],
        "xyz",
    ),
    # Matching with multiple partially overlapping filters
    (
        "qaqxqbyqzq",
        [(2, 6), (7, 8)],
        "qayzq",
    ),
    # Matching with multiple fully overlapping filters
    (
        "qaqxqybqzq",
        [(2, 8)],
        "qazq",
    ),
    # Matching with and without groups, with single dominated match
    (
        "# asdasdasdasdsab",
        [(0, 17)],
        "",
    ),
    # Matching with and without groups, with partially overlapping filters
    (
        "/*a*/ub",
        [(0, 6)],
        "b",
    ),
    # Non-matching with groups
    (
        "xasdyasdx",
        [],
        "xasdyasdx",
    ),
    # Multiple lines with non-overlapping filters
    (
        "#ab\na2b",
        [(0, 3), (5, 6)],
        "\nab",
    ),
    # CVS keyword
    (
        "$Author: John Doe $",
        [(8, 18)],
        "$Author:$",
    ),

])
def test_filter_text(text, ignored_ranges, expected_text):
    from meld.filediff import FileDiff
    from meld.filters import FilterEntry

    filter_patterns = [
        '#.*',
        r'/\*.*\*/',
        'a(.*)b',
        'x(.*)y(.*)z',
        r'\$\w+:([^\n$]+)\$'
    ]
    filters = [
        FilterEntry.new_from_gsetting(("name", True, f), FilterEntry.REGEX)
        for f in filter_patterns
    ]

    filediff = mock.MagicMock()
    filediff.text_filters = filters
    filter_text = FileDiff._filter_text

    buf = Gtk.TextBuffer()
    buf.create_tag("inline")
    buf.create_tag("dimmed")
    buf.set_text(text)
    start, end = buf.get_bounds()

    text = filter_text(
        filediff, buf.get_text(start, end, False), buf, start, end)

    # Find ignored ranges
    tag = buf.get_tag_table().lookup("dimmed")
    toggles = []
    it = start.copy()
    if it.toggles_tag(tag):
        toggles.append(it.get_offset())
    while it.forward_to_tag_toggle(tag):
        toggles.append(it.get_offset())
    toggles = list(zip(toggles[::2], toggles[1::2]))

    print("Text:", text)
    print("Toggles:", toggles)

    assert toggles == ignored_ranges
    assert text == expected_text




from unittest import mock

import pytest

from meld.const import ActionMode
from meld.matchers.myers import DiffChunk


def make_chunk(chunk_type):
    return DiffChunk(chunk_type, 0, 1, 0, 1)


@pytest.mark.parametrize("mode, editable, chunk, expected_action", [
    # Replace mode with replace chunks
    (ActionMode.Replace, (True, True), make_chunk('replace'), ActionMode.Replace),
    (ActionMode.Replace, (True, False), make_chunk('replace'), ActionMode.Delete),
    (ActionMode.Replace, (False, True), make_chunk('replace'), ActionMode.Replace),
    (ActionMode.Replace, (False, False), make_chunk('replace'), None),
    # Replace mode with delete chunks
    (ActionMode.Replace, (True, True), make_chunk('delete'), ActionMode.Replace),
    (ActionMode.Replace, (True, False), make_chunk('delete'), ActionMode.Delete),
    (ActionMode.Replace, (False, True), make_chunk('delete'), ActionMode.Replace),
    (ActionMode.Replace, (False, False), make_chunk('delete'), None),
    # Delete mode makes a slightly weird choice to remove non-delete
    # actions while in delete mode; insert mode makes the opposite
    # choice
    #
    # Delete mode with replace chunks
    (ActionMode.Delete, (True, True), make_chunk('replace'), ActionMode.Delete),
    (ActionMode.Delete, (True, False), make_chunk('replace'), ActionMode.Delete),
    (ActionMode.Delete, (False, True), make_chunk('replace'), None),
    (ActionMode.Delete, (False, False), make_chunk('replace'), None),
    # Delete mode with delete chunks
    (ActionMode.Delete, (True, True), make_chunk('delete'), ActionMode.Delete),
    (ActionMode.Delete, (True, False), make_chunk('delete'), ActionMode.Delete),
    (ActionMode.Delete, (False, True), make_chunk('delete'), None),
    (ActionMode.Delete, (False, False), make_chunk('delete'), None),
    # Insert mode with replace chunks
    (ActionMode.Insert, (True, True), make_chunk('replace'), ActionMode.Insert),
    (ActionMode.Insert, (True, False), make_chunk('replace'), ActionMode.Delete),
    (ActionMode.Insert, (False, True), make_chunk('replace'), ActionMode.Insert),
    (ActionMode.Insert, (False, False), make_chunk('replace'), None),
    # Insert mode with delete chunks
    (ActionMode.Insert, (True, True), make_chunk('delete'), ActionMode.Replace),
    (ActionMode.Insert, (True, False), make_chunk('delete'), ActionMode.Delete),
    (ActionMode.Insert, (False, True), make_chunk('delete'), ActionMode.Replace),
    (ActionMode.Insert, (False, False), make_chunk('delete'), None),
    # We should never have insert chunks here
    (ActionMode.Replace, (True, True), make_chunk('insert'), None),
    (ActionMode.Replace, (True, False), make_chunk('insert'), None),
    (ActionMode.Replace, (False, True), make_chunk('insert'), None),
    (ActionMode.Replace, (False, False), make_chunk('insert'), None),

    # TODO: Add tests for conflict chunks
])
def test_classify_change_actions(mode, editable, chunk, expected_action):

    # These tests are disabled due to a segfault on the CI machines.
    return

    from meld.actiongutter import ActionGutter

    source_editable, target_editable = editable

    with mock.patch.object(ActionGutter, 'icon_direction'):
        renderer = ActionGutter()
        renderer._source_view = mock.Mock()
        renderer._source_view.get_editable.return_value = source_editable
        renderer._target_view = mock.Mock()
        renderer._target_view.get_editable.return_value = target_editable
        renderer.action_mode = mode

        action = renderer._classify_change_actions(chunk)
        assert action == expected_action



#!/usr/bin/env python3

import os
import sys
from compileall import compile_dir
from os import environ, path
from subprocess import call

if not environ.get('DESTDIR', ''):
    PREFIX = environ.get('MESON_INSTALL_PREFIX', '/usr/local')
    DATA_DIR = path.join(PREFIX, 'share')
    print('Updating icon cache...')
    call(['gtk-update-icon-cache', '-qtf', path.join(DATA_DIR, 'icons', 'hicolor')])
    print("Compiling new schemas")
    call(["glib-compile-schemas", path.join(DATA_DIR, 'glib-2.0', 'schemas')])
    print("Updating desktop database")
    call(["update-desktop-database", path.join(DATA_DIR, 'applications')])

# Byte-compilation is enabled by passing the site-packages path to this script
if len(sys.argv) > 1:
    print('Byte-compiling Python module...')
    destdir = os.getenv("DESTDIR", "")
    python_source_install_path = sys.argv[1]
    if destdir:
        # The install path here will be absolute, so we can't use join()
        install_path = destdir + os.path.sep + python_source_install_path
    else:
        install_path = python_source_install_path
    compile_dir(os.path.join(install_path, "meld"), optimize=1)



# Copyright (C) 2016 Kai Willadsen <kai.willadsen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime

from gi.repository import GObject, Gtk


class CellRendererDate(Gtk.CellRendererText):

    __gtype_name__ = "CellRendererDate"

    #: We use negative 32-bit Unix timestamp to threshold our valid values
    MIN_TIMESTAMP = -2147483648
    DATETIME_FORMAT = "%a %d %b %Y %H:%M:%S"

    def _format_datetime(self, dt: datetime.datetime) -> str:
        return dt.strftime(self.DATETIME_FORMAT)

    def get_timestamp(self):
        return getattr(self, '_datetime', self.MIN_TIMESTAMP)

    def set_timestamp(self, value):
        if value == self.get_timestamp():
            return
        if value <= self.MIN_TIMESTAMP:
            time_str = ''
        else:
            try:
                mod_datetime = datetime.datetime.fromtimestamp(value)
                time_str = self._format_datetime(mod_datetime)
            except Exception:
                time_str = ''
        self.props.markup = time_str
        self._datetime = value

    timestamp = GObject.Property(
        type=float,
        nick="Unix timestamp to display",
        getter=get_timestamp,
        setter=set_timestamp,
    )


class CellRendererISODate(CellRendererDate):

    __gtype_name__ = "CellRendererISODate"

    def _format_datetime(self, dt: datetime.datetime) -> str:
        # Limit our ISO display to seconds (i.e., no milli or
        # microseconds) for usability
        return dt.isoformat(timespec="seconds")


class CellRendererByteSize(Gtk.CellRendererText):

    __gtype_name__ = "CellRendererByteSize"

    def get_bytesize(self):
        return getattr(self, '_bytesize', -1)

    def set_bytesize(self, value):
        if value == self.get_bytesize():
            return
        if value == -1:
            byte_str = ''
        else:
            suffixes = (
                'B', 'kB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB'
            )
            size = float(value)
            unit = 0
            while size > 1000 and unit < len(suffixes) - 1:
                size /= 1000
                unit += 1
            format_str = "%.1f %s" if unit > 0 else "%d %s"
            byte_str = format_str % (size, suffixes[unit])
        self.props.markup = byte_str
        self._bytesize = value

    bytesize = GObject.Property(
        type=GObject.TYPE_INT64,
        nick="Byte size to display",
        getter=get_bytesize,
        setter=set_bytesize,
    )


class CellRendererFileMode(Gtk.CellRendererText):

    __gtype_name__ = "CellRendererFileMode"

    def get_file_mode(self):
        return getattr(self, '_file_mode', -1)

    def set_file_mode(self, value):
        if value == self.get_file_mode():
            return
        if value == -1.0:
            mode_str = ''
        else:
            perms = []
            rwx = ((4, 'r'), (2, 'w'), (1, 'x'))
            for group_index in (6, 3, 0):
                group = value >> group_index & 7
                perms.extend([p if group & i else '-' for i, p in rwx])
            mode_str = "".join(perms)
        self.props.markup = mode_str
        self._file_mode = value

    file_mode = GObject.Property(
        type=int,
        nick="Byte size to display",
        getter=get_file_mode,
        setter=set_file_mode,
    )



# Copyright (C) 2002-2006 Stephen Kennedy <stevek@gnome.org>
# Copyright (C) 2011-2013 Kai Willadsen <kai.willadsen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import enum
import logging
import os
import pipes
import shlex
import string
import subprocess
import sys
from typing import List, Optional, Sequence

from gi.repository import Gdk, Gio, GLib, GObject, Gtk

from meld.conf import _
from meld.recent import RecentType
from meld.settings import settings
from meld.task import FifoScheduler

log = logging.getLogger(__name__)


def make_custom_editor_command(path: str, line: int = 0) -> Sequence[str]:
    custom_command = settings.get_string('custom-editor-command')
    fmt = string.Formatter()
    replacements = [tok[1] for tok in fmt.parse(custom_command)]

    if not any(replacements):
        return [custom_command, path]
    elif not all(r in (None, 'file', 'line') for r in replacements):
        log.error("Unsupported fields found")
        return [custom_command, path]
    else:
        cmd = custom_command.format(file=pipes.quote(path), line=line)
    return shlex.split(cmd)


def open_files_external(
        paths: Optional[List[str]] = None, *,
        gfiles: Optional[List[Gio.File]] = None,
        line: int = 0) -> None:

    def os_open(path: str, uri: str):
        if not path:
            return
        if sys.platform == "win32":
            os.startfile(path)
        elif sys.platform == "darwin":
            subprocess.Popen(["open", path])
        else:
            Gtk.show_uri(
                Gdk.Screen.get_default(), uri, Gtk.get_current_event_time())

    def open_cb(source, result, *data):
        info = source.query_info_finish(result)
        file_type = info.get_file_type()
        path, uri = source.get_path(), source.get_uri()
        if file_type == Gio.FileType.DIRECTORY:
            os_open(path, uri)
        elif file_type == Gio.FileType.REGULAR:
            content_type = info.get_content_type()
            # FIXME: Content types are broken on Windows with current gio
            # If we can't access a content type, assume it's text.
            if not content_type or Gio.content_type_is_a(
                    content_type, "text/plain"):
                if settings.get_boolean('use-system-editor'):
                    gfile = Gio.File.new_for_path(path)
                    if sys.platform == "win32":
                        handler = gfile.query_default_handler(None)
                        result = handler.launch([gfile], None)
                    else:
                        uri = gfile.get_uri()
                        Gio.AppInfo.launch_default_for_uri(
                            uri, None)
                else:
                    editor = make_custom_editor_command(path, line)
                    if editor:
                        # TODO: If the editor is badly set up, this fails
                        # silently
                        subprocess.Popen(editor)
                    else:
                        os_open(path, uri)
            else:
                os_open(path, uri)
        else:
            # TODO: Add some kind of 'failed to open' notification
            pass

    query_attrs = ",".join((Gio.FILE_ATTRIBUTE_STANDARD_TYPE,
                            Gio.FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE))

    if not gfiles:
        gfiles = [Gio.File.new_for_path(s) for s in paths]

    for f in gfiles:
        f.query_info_async(
            query_attrs, Gio.FileQueryInfoFlags.NONE, GLib.PRIORITY_LOW,
            None, open_cb, None)


class ComparisonState(enum.IntEnum):
    # TODO: Consider use-cases for states in gedit-enum-types.c
    Normal = 0
    Closing = 1
    SavingError = 2


class LabeledObjectMixin(GObject.GObject):

    label_text = _("untitled")
    tooltip_text = None

    @GObject.Signal
    def label_changed(self, label_text: str, tooltip_text: str) -> None:
        ...


class MeldDoc(LabeledObjectMixin, GObject.GObject):
    """Base class for documents in the meld application.
    """

    @GObject.Signal(name='close')
    def close_signal(self, exit_code: int) -> None:
        ...

    @GObject.Signal(name='create-diff')
    def create_diff_signal(
            self, gfiles: object, options: object) -> None:
        ...

    @GObject.Signal('file-changed')
    def file_changed_signal(self, path: str) -> None:
        ...

    @GObject.Signal
    def tab_state_changed(self, old_state: int, new_state: int) -> None:
        ...

    @GObject.Signal(
        name='move-diff',
        flags=GObject.SignalFlags.RUN_FIRST | GObject.SignalFlags.ACTION,
    )
    def move_diff(self, direction: int) -> None:
        self.next_diff(direction)

    def __init__(self) -> None:
        super().__init__()
        self.scheduler = FifoScheduler()
        self.num_panes = 0
        self.view_action_group = Gio.SimpleActionGroup()
        self._state = ComparisonState.Normal

    @property
    def state(self) -> ComparisonState:
        return self._state

    @state.setter
    def state(self, value: ComparisonState) -> None:
        if value == self._state:
            return
        self.tab_state_changed.emit(self._state, value)
        self._state = value

    def get_comparison(self) -> RecentType:
        """Get the comparison type and URI(s) being compared"""
        pass

    def action_stop(self, *args) -> None:
        if self.scheduler.tasks_pending():
            self.scheduler.remove_task(self.scheduler.get_current_task())

    def on_file_changed(self, filename: str):
        pass

    def set_labels(self, lst: Sequence[str]) -> None:
        pass

    def get_action_state(self, action_name: str):
        action = self.view_action_group.lookup_action(action_name)
        if not action:
            log.error(f'No action {action_name!r} found')
            return
        return action.get_state().unpack()

    def set_action_state(self, action_name: str, state) -> None:
        # TODO: Try to do GLib.Variant things here instead of in callers
        action = self.view_action_group.lookup_action(action_name)
        if not action:
            log.error(f'No action {action_name!r} found')
            return
        action.set_state(state)

    def set_action_enabled(self, action_name: str, enabled: bool) -> None:
        action = self.view_action_group.lookup_action(action_name)
        if not action:
            log.error(f'No action {action_name!r} found')
            return
        action.set_enabled(enabled)

    def on_container_switch_in_event(self, window):
        """Called when the container app switches to this tab"""

        window.insert_action_group(
            'view', getattr(self, 'view_action_group', None))

        if hasattr(self, "get_filter_visibility"):
            text, folder, vc = self.get_filter_visibility()
        else:
            text, folder, vc = False, False, False

        if hasattr(self, "get_conflict_visibility"):
            show_conflict_actions = self.get_conflict_visibility()
        else:
            show_conflict_actions = False

        window.text_filter_button.set_visible(text)
        window.folder_filter_button.set_visible(folder)
        window.vc_filter_button.set_visible(vc)

        window.next_conflict_button.set_visible(show_conflict_actions)
        window.previous_conflict_button.set_visible(show_conflict_actions)

        if hasattr(self, "focus_pane") and self.focus_pane:
            self.scheduler.add_task(self.focus_pane.grab_focus)

    def on_container_switch_out_event(self, window):
        """Called when the container app switches away from this tab"""

        window.insert_action_group('view', None)

    # FIXME: Here and in subclasses, on_delete_event are not real GTK+
    # event handlers, and should be renamed.
    def on_delete_event(self) -> Gtk.ResponseType:
        """Called when the docs container is about to close.

        A doc normally returns Gtk.ResponseType.OK, but may instead return
        Gtk.ResponseType.CANCEL to request that the container not delete it.
        """
        return Gtk.ResponseType.OK



# Copyright (C) 2012-2013 Kai Willadsen <kai.willadsen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GObject, Gtk, GtkSource, Pango

from meld.conf import _
from meld.ui.bufferselectors import EncodingSelector, SourceLangSelector


class MeldStatusMenuButton(Gtk.MenuButton):
    """Compact menu button with arrow indicator for use in a status bar

    Implementation based on gedit-status-menu-button.c
    Copyright (C) 2008 - Jesse van den Kieboom
    """

    __gtype_name__ = "MeldStatusMenuButton"

    style = b"""
    * {
      padding: 1px 8px 2px 4px;
      border: 0;
      outline-width: 0;
    }
    """

    css_provider = Gtk.CssProvider()
    css_provider.load_from_data(style)

    def get_label(self):
        return self._label.get_text()

    def set_label(self, markup):
        if markup == self._label.get_text():
            return
        self._label.set_markup(markup)

    label = GObject.Property(
        type=str,
        default=None,
        getter=get_label,
        setter=set_label,
    )

    def __init__(self):
        super().__init__()

        style_context = self.get_style_context()
        style_context.add_provider(
            self.css_provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
        style_context.add_class('flat')

        # Ideally this would be a template child, but there's still no
        # Python support for this.
        label = Gtk.Label(
            single_line_mode=True,
            halign=Gtk.Align.START,
            valign=Gtk.Align.BASELINE,
            xalign=1.0,
            ellipsize=Pango.EllipsizeMode.END,
        )

        arrow = Gtk.Image.new_from_icon_name(
            'pan-down-symbolic', Gtk.IconSize.SMALL_TOOLBAR)
        arrow.props.valign = Gtk.Align.BASELINE

        box = Gtk.Box()
        box.set_spacing(3)
        box.add(label)
        box.add(arrow)
        box.show_all()

        self.remove(self.get_child())
        self.add(box)

        self._label = label

    def set_label_width(self, width):
        self._label.set_width_chars(width)


class MeldStatusBar(Gtk.Statusbar):
    __gtype_name__ = "MeldStatusBar"

    __gsignals__ = {
        'start-go-to-line': (
            GObject.SignalFlags.ACTION, None, tuple()),
        'go-to-line': (
            GObject.SignalFlags.RUN_FIRST, None, (int,)),
        'encoding-changed': (
            GObject.SignalFlags.RUN_FIRST, None, (GtkSource.Encoding,)),
    }

    cursor_position = GObject.Property(
        type=object,
        nick="The position of the cursor displayed in the status bar",
        default=None,
    )

    source_encoding = GObject.Property(
        type=GtkSource.Encoding,
        nick="The file encoding displayed in the status bar",
        default=GtkSource.Encoding.get_utf8(),
    )

    source_language = GObject.Property(
        type=GtkSource.Language,
        nick="The GtkSourceLanguage displayed in the status bar",
        default=None,
    )

    # Abbreviation for line, column so that it will fit in the status bar
    _line_column_text = _("Ln {line}, Col {column}")

    def __init__(self):
        super().__init__()
        self.props.margin = 0
        self.props.spacing = 6

        hbox = self.get_message_area()
        label = hbox.get_children()[0]
        hbox.props.spacing = 6
        label.props.ellipsize = Pango.EllipsizeMode.NONE
        hbox.remove(label)
        hbox.pack_end(label, False, True, 0)

    def do_realize(self):
        Gtk.Statusbar.do_realize(self)

        self.box_box = Gtk.HBox(homogeneous=False, spacing=6)
        self.pack_end(self.box_box, False, True, 0)
        self.box_box.pack_end(
            self.construct_line_display(), False, True, 0)
        self.box_box.pack_end(
            self.construct_highlighting_selector(), False, True, 0)
        self.box_box.pack_end(
            self.construct_encoding_selector(), False, True, 0)
        self.box_box.pack_end(
            self.construct_display_popover(), False, True, 0)
        self.box_box.show_all()

    def construct_line_display(self):

        # Note that we're receiving one-based line numbers from the
        # user and storing and emitting zero-base line numbers.

        def go_to_line_text(text):
            try:
                line = int(text)
            except ValueError:
                return
            self.emit('go-to-line', max(0, line - 1))

        def line_entry_mapped(entry):
            line, offset = self.props.cursor_position
            entry.set_text(str(line + 1))

        # This handler causes a failed assertion due to the `position`
        # out param (see pygobject#12), but we don't need it here.
        def line_entry_insert_text(entry, new_text, length, position):
            if not new_text.isdigit():
                GObject.signal_stop_emission_by_name(entry, 'insert-text')
                return

        def line_entry_changed(entry):
            go_to_line_text(entry.get_text())

        def line_entry_activated(entry):
            go_to_line_text(entry.get_text())
            pop.popdown()

        entry = Gtk.Entry()
        entry.set_tooltip_text(_('Line you want to move the cursor to'))
        entry.set_icon_from_icon_name(
            Gtk.EntryIconPosition.PRIMARY, 'go-jump-symbolic')
        entry.set_icon_activatable(Gtk.EntryIconPosition.PRIMARY, False)
        entry.set_input_purpose(Gtk.InputPurpose.DIGITS)
        entry.connect('map', line_entry_mapped)
        entry.connect('insert-text', line_entry_insert_text)
        entry.connect('changed', line_entry_changed)
        entry.connect('activate', line_entry_activated)

        selector = Gtk.Grid()
        selector.set_border_width(6)
        selector.add(entry)
        selector.show_all()

        pop = Gtk.Popover()
        pop.set_position(Gtk.PositionType.TOP)
        pop.add(selector)

        def format_cursor_position(binding, cursor):
            line, offset = cursor
            return self._line_column_text.format(
                line=line + 1, column=offset + 1)

        button = MeldStatusMenuButton()
        self.bind_property(
            'cursor_position', button, 'label', GObject.BindingFlags.DEFAULT,
            format_cursor_position)
        self.connect('start-go-to-line', lambda *args: button.clicked())
        button.set_popover(pop)
        # Set a label width to avoid other widgets moving on cursor change
        reasonable_width = len(format_cursor_position(None, (1000, 100))) - 2
        button.set_label_width(reasonable_width)
        button.show()

        return button

    def construct_encoding_selector(self):
        def change_encoding(selector, encoding):
            self.emit('encoding-changed', encoding)
            pop.hide()

        def set_initial_encoding(selector):
            selector.select_value(self.props.source_encoding)

        selector = EncodingSelector()
        selector.connect('encoding-selected', change_encoding)
        selector.connect('map', set_initial_encoding)

        pop = Gtk.Popover()
        pop.set_position(Gtk.PositionType.TOP)
        pop.add(selector)

        button = MeldStatusMenuButton()
        self.bind_property(
            'source-encoding', button, 'label',
            GObject.BindingFlags.DEFAULT | GObject.BindingFlags.SYNC_CREATE,
            lambda binding, enc: selector.get_value_label(enc))
        button.set_popover(pop)
        button.show()

        return button

    def construct_highlighting_selector(self):
        def change_language(selector, lang):
            # TODO: Our other GObject properties are expected to be
            # updated through a bound state from our parent. This is
            # the only place where we assign to them instead of
            # emitting a signal, and it makes the class logic as a
            # whole kind of confusing.
            self.props.source_language = lang
            pop.hide()

        def set_initial_language(selector):
            selector.select_value(self.props.source_language)

        selector = SourceLangSelector()
        selector.connect('language-selected', change_language)
        selector.connect('map', set_initial_language)

        pop = Gtk.Popover()
        pop.set_position(Gtk.PositionType.TOP)
        pop.add(selector)

        button = MeldStatusMenuButton()
        self.bind_property(
            'source-language', button, 'label',
            GObject.BindingFlags.DEFAULT | GObject.BindingFlags.SYNC_CREATE,
            lambda binding, enc: selector.get_value_label(enc))
        button.set_popover(pop)
        button.show()

        return button

    def construct_display_popover(self):
        builder = Gtk.Builder.new_from_resource(
            '/org/gnome/meld/ui/statusbar-menu.ui')
        menu = builder.get_object('statusbar-menu')

        pop = Gtk.Popover()
        pop.bind_model(menu, 'view-local')
        pop.set_position(Gtk.PositionType.TOP)

        button = MeldStatusMenuButton()
        # TRANSLATORS: This is the status bar label for a group of settings,
        # such as text wrapping, show line numbers, whitespace, etc.
        button.set_label(_('Display'))
        button.set_popover(pop)
        button.show()

        return button



# Copyright (C) 2002-2005 Stephen Kennedy <stevek@gnome.org>
# Copyright (C) 2012-2015 Kai Willadsen <kai.willadsen@gmail.com>

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import shutil
import tempfile

from meld.conf import _
from meld.vc import _vc


class Vc(_vc.Vc):

    CMD = None
    # Translators: This is the displayed name of a version control system
    # when no version control system is actually found.
    NAME = _("None")
    VC_DIR = "."

    def _update_tree_state_cache(*args):
        pass

    def get_path_for_repo_file(self, path, commit=None):
        suffix = os.path.splitext(path)[1]
        with tempfile.NamedTemporaryFile(
                prefix='meld-tmp', suffix=suffix, delete=False) as f:
            with open(path, 'rb') as vc_file:
                shutil.copyfileobj(vc_file, f)
        return f.name



# Copyright (C) 2002-2006 Stephen Kennedy <stevek@gnome.org>
# Copyright (C) 2009-2013 Kai Willadsen <kai.willadsen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from typing import Any, List, Optional

from gi.repository import Gio, GLib, GObject, GtkSource

from meld.conf import _
from meld.settings import bind_settings

log = logging.getLogger(__name__)


class MeldBuffer(GtkSource.Buffer):

    __gtype_name__ = "MeldBuffer"

    __gsettings_bindings__ = (
        ('highlight-syntax', 'highlight-syntax'),
    )

    def __init__(self):
        super().__init__()
        bind_settings(self)
        self.data = MeldBufferData()
        self.undo_sequence = None

    def do_begin_user_action(self, *args):
        if self.undo_sequence:
            self.undo_sequence.begin_group()

    def do_end_user_action(self, *args):
        if self.undo_sequence:
            self.undo_sequence.end_group()

    def get_iter_at_line_or_eof(self, line):
        """Return a Gtk.TextIter at the given line, or the end of the buffer.

        This method is like get_iter_at_line, but if asked for a position past
        the end of the buffer, this returns the end of the buffer; the
        get_iter_at_line behaviour is to return the start of the last line in
        the buffer.
        """
        if line >= self.get_line_count():
            return self.get_end_iter()
        return self.get_iter_at_line(line)

    def insert_at_line(self, line, text):
        """Insert text at the given line, or the end of the buffer.

        This method is like insert, but if asked to insert something past the
        last line in the buffer, this will insert at the end, and will add a
        linebreak before the inserted text. The last line in a Gtk.TextBuffer
        is guaranteed never to have a newline, so we need to handle this.
        """
        if line >= self.get_line_count():
            # TODO: We need to insert a linebreak here, but there is no
            # way to be certain what kind of linebreak to use.
            text = "\n" + text
        it = self.get_iter_at_line_or_eof(line)
        self.insert(it, text)
        return it


class MeldBufferData(GObject.GObject):

    @GObject.Signal('file-changed')
    def file_changed_signal(self) -> None:
        ...

    encoding = GObject.Property(
        type=GtkSource.Encoding,
        nick="The file encoding of the linked GtkSourceFile",
        default=GtkSource.Encoding.get_utf8(),
    )

    def __init__(self):
        super().__init__()
        self._gfile = None
        self._label = None
        self._monitor = None
        self._sourcefile = None
        self.reset(gfile=None)

    def reset(self, gfile):
        same_file = gfile and self._gfile and gfile.equal(self._gfile)
        self.gfile = gfile
        if same_file:
            self.label = self._label
        else:
            self.label = gfile.get_parse_name() if gfile else None
        self.loaded = False
        self.savefile = None

    def __del__(self):
        self.disconnect_monitor()

    @property
    def label(self):
        # TRANSLATORS: This is the label of a new, currently-unnamed file.
        return self._label or _("<unnamed>")

    @label.setter
    def label(self, value):
        if not value:
            return
        if not isinstance(value, str):
            log.warning('Invalid label ignored "%r"', value)
            return
        self._label = value

    def connect_monitor(self):
        if not self._gfile:
            return
        monitor = self._gfile.monitor_file(Gio.FileMonitorFlags.NONE, None)
        handler_id = monitor.connect('changed', self._handle_file_change)
        self._monitor = monitor, handler_id

    def disconnect_monitor(self):
        if not self._monitor:
            return
        monitor, handler_id = self._monitor
        monitor.disconnect(handler_id)
        monitor.cancel()
        self._monitor = None

    def _query_mtime(self, gfile):
        try:
            time_query = ",".join((Gio.FILE_ATTRIBUTE_TIME_MODIFIED,
                                   Gio.FILE_ATTRIBUTE_TIME_MODIFIED_USEC))
            info = gfile.query_info(time_query, 0, None)
        except GLib.GError:
            return None
        mtime = info.get_modification_time()
        return (mtime.tv_sec, mtime.tv_usec)

    def _handle_file_change(self, monitor, f, other_file, event_type):
        mtime = self._query_mtime(f)
        if self._disk_mtime and mtime and mtime > self._disk_mtime:
            self.file_changed_signal.emit()
        self._disk_mtime = mtime or self._disk_mtime

    @property
    def gfile(self):
        return self._gfile

    @gfile.setter
    def gfile(self, value):
        self.disconnect_monitor()
        self._gfile = value
        self._sourcefile = GtkSource.File()
        self._sourcefile.set_location(value)
        self._sourcefile.bind_property(
            'encoding', self, 'encoding', GObject.BindingFlags.DEFAULT)

        self.update_mtime()
        self.connect_monitor()

    @property
    def sourcefile(self):
        return self._sourcefile

    @property
    def gfiletarget(self):
        return self.savefile or self.gfile

    @property
    def is_special(self):
        try:
            info = self._gfile.query_info(
                Gio.FILE_ATTRIBUTE_STANDARD_TYPE, 0, None)
            return info.get_file_type() == Gio.FileType.SPECIAL
        except (AttributeError, GLib.GError):
            return False

    @property
    def writable(self):
        try:
            info = self.gfiletarget.query_info(
                Gio.FILE_ATTRIBUTE_ACCESS_CAN_WRITE, 0, None)
        except GLib.GError as err:
            if err.code == Gio.IOErrorEnum.NOT_FOUND:
                return True
            return False
        except AttributeError:
            return False
        return info.get_attribute_boolean(Gio.FILE_ATTRIBUTE_ACCESS_CAN_WRITE)

    def update_mtime(self):
        if self._gfile:
            self._disk_mtime = self._query_mtime(self._gfile)
            self._mtime = self._disk_mtime

    def current_on_disk(self):
        return self._mtime == self._disk_mtime


class BufferLines:
    """Gtk.TextBuffer shim with line-based access and optional filtering

    This class allows a Gtk.TextBuffer to be treated as a list of lines of
    possibly-filtered text. If no filter is given, the raw output from the
    Gtk.TextBuffer is used.
    """

    #: Cached copy of the (possibly filtered) text in a single line,
    #: where an entry of None indicates that there is no cached result
    #: available.
    lines: List[Optional[str]]

    def __init__(self, buf, textfilter=None, *, cache_debug: bool = False):
        self.buf = buf
        if textfilter is not None:
            self.textfilter = textfilter
        else:
            self.textfilter = lambda x, buf, start_iter, end_iter: x

        self.lines = [None] * self.buf.get_line_count()
        self.mark = buf.create_mark(
            "bufferlines-insert", buf.get_start_iter(), True,
        )

        buf.connect("insert-text", self.on_insert_text),
        buf.connect("delete-range", self.on_delete_range),
        buf.connect_after("insert-text", self.after_insert_text),
        if cache_debug:
            buf.connect_after("insert-text", self._check_cache_invariant)
            buf.connect_after("delete-range", self._check_cache_invariant)

    def _check_cache_invariant(self, *args: Any) -> None:
        if len(self.lines) != len(self):
            log.error(
                "Cache line count does not match buffer line count: "
                f"{len(self.lines)} != {len(self)}",
            )

    def clear_cache(self) -> None:
        self.lines = [None] * self.buf.get_line_count()

    def on_insert_text(self, buf, it, text, textlen):
        buf.move_mark(self.mark, it)

    def after_insert_text(self, buf, it, newtext, textlen):
        start_idx = buf.get_iter_at_mark(self.mark).get_line()
        end_idx = it.get_line() + 1
        # Replace the insertion-point cache line with a list of empty
        # lines. In the single-line case this will be a single element
        # substitution; for multi-line inserts, we will replace the
        # single insertion point line with several empty cache lines.
        self.lines[start_idx:start_idx + 1] = [None] * (end_idx - start_idx)

    def on_delete_range(self, buf, it0, it1):
        start_idx = it0.get_line()
        end_idx = it1.get_line() + 1
        self.lines[start_idx:end_idx] = [None]

    def __getitem__(self, key):
        if isinstance(key, slice):
            lo, hi, _ = key.indices(self.buf.get_line_count())

            for idx in range(lo, hi):
                if self.lines[idx] is None:
                    self.lines[idx] = self[idx]

            return self.lines[lo:hi]

        elif isinstance(key, int):
            if key >= len(self):
                raise IndexError

            if self.lines[key] is None:
                line_start = self.buf.get_iter_at_line_or_eof(key)
                line_end = line_start.copy()
                if not line_end.ends_line():
                    line_end.forward_to_line_end()
                txt = self.buf.get_text(line_start, line_end, False)
                txt = self.textfilter(txt, self.buf, line_start, line_end)
                self.lines[key] = txt

            return self.lines[key]

    def __len__(self):
        return self.buf.get_line_count()


class BufferAction:
    """A helper to undo/redo text insertion/deletion into/from a text buffer"""

    def __init__(self, buf, offset, text):
        self.buffer = buf
        self.offset = offset
        self.text = text

    def delete(self):
        start = self.buffer.get_iter_at_offset(self.offset)
        end = self.buffer.get_iter_at_offset(self.offset + len(self.text))
        self.buffer.delete(start, end)
        self.buffer.place_cursor(end)
        return [self]

    def insert(self):
        start = self.buffer.get_iter_at_offset(self.offset)
        self.buffer.place_cursor(start)
        self.buffer.insert(start, self.text)
        return [self]


class BufferInsertionAction(BufferAction):
    undo = BufferAction.delete
    redo = BufferAction.insert


class BufferDeletionAction(BufferAction):
    undo = BufferAction.insert
    redo = BufferAction.delete




from typing import Dict, Sequence, Union

from gi.repository import Gtk

VIEW_ACCELERATORS: Dict[str, Union[str, Sequence[str]]] = {
    'app.quit': '<Primary>Q',
    'view.find': '<Primary>F',
    'view.find-next': ('<Primary>G', 'F3'),
    'view.find-previous': ('<Primary><Shift>G', '<Shift>F3'),
    'view.find-replace': '<Primary>H',
    'view.go-to-line': '<Primary>I',
    # Overridden in CSS
    'view.next-change': ('<Alt>Down', '<Alt>KP_Down', '<Primary>D'),
    'view.next-pane': '<Alt>Page_Down',
    'view.open-external': '<Primary><Shift>O',
    # Overridden in CSS
    'view.previous-change': ('<Alt>Up', '<Alt>KP_Up', '<Primary>E'),
    'view.previous-pane': '<Alt>Page_Up',
    'view.redo': '<Primary><Shift>Z',
    'view.refresh': ('<control>R', 'F5'),
    'view.save': '<Primary>S',
    'view.save-all': '<Primary><Shift>L',
    'view.save-as': '<Primary><Shift>S',
    'view.undo': '<Primary>Z',
    'win.close': '<Primary>W',
    'win.gear-menu': 'F10',
    'win.new-tab': '<Primary>N',
    'win.stop': 'Escape',
    # File comparison actions
    'view.file-previous-conflict': '<Primary>J',
    'view.file-next-conflict': '<Primary>K',
    'view.file-push-left': '<Alt>Left',
    'view.file-push-right': '<Alt>Right',
    'view.file-pull-left': '<Alt><shift>Right',
    'view.file-pull-right': '<Alt><shift>Left',
    'view.file-copy-left-up': '<Alt>bracketleft',
    'view.file-copy-right-up': '<Alt>bracketright',
    'view.file-copy-left-down': '<Alt>semicolon',
    'view.file-copy-right-down': '<Alt>quoteright',
    'view.file-delete': ('<Alt>Delete', '<Alt>KP_Delete'),
    'view.show-overview-map': 'F9',
    # Folder comparison actions
    'view.folder-compare': 'Return',
    'view.folder-copy-left': '<Alt>Left',
    'view.folder-copy-right': '<Alt>Right',
    'view.folder-delete': 'Delete',
    # Version control actions
    'view.vc-commit': '<Primary>M',
    'view.vc-console-visible': 'F9',
    # Swap the two panes
    'view.swap-2-panes': '<Alt>backslash',
}


def register_accels(app: Gtk.Application):
    for name, accel in VIEW_ACCELERATORS.items():
        accel = accel if isinstance(accel, tuple) else (accel,)
        app.set_accels_for_action(name, accel)




# This file exists so that it can be referenced as an external Glade
# module, allowing editing of our UI files without having widgets that
# can't be instantiated everywhere.

from meld import actiongutter  # noqa: F401
from meld import chunkmap  # noqa: F401
from meld import diffgrid  # noqa: F401
from meld import linkmap  # noqa: F401
from meld import preferences  # noqa: F401
from meld import sourceview  # noqa: F401
from meld.ui import emblemcellrenderer  # noqa: F401
from meld.ui import filebutton  # noqa: F401
from meld.ui import historyentry  # noqa: F401
from meld.ui import msgarea  # noqa: F401
from meld.ui import notebook  # noqa: F401
from meld.ui import pathlabel  # noqa: F401
from meld.ui import recentselector  # noqa: F401
from meld.ui import statusbar  # noqa: F401




TEST_REQUIRES = {
    "GLib": "2.0",
    "Gtk": "3.0",
    "GtkSource": "4",
}


def enforce_requires():
    import gi

    for namespace, version in TEST_REQUIRES.items():
        gi.require_version(namespace, version)


enforce_requires()



# Copyright (C) 2002-2006 Stephen Kennedy <stevek@gnome.org>
# Copyright (C) 2009-2010, 2013 Kai Willadsen <kai.willadsen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import difflib
import os

from gi.repository import Gdk, Gio, GLib, Gtk, GtkSource

from meld.conf import _
from meld.iohelpers import prompt_save_filename
from meld.misc import error_dialog
from meld.settings import get_meld_settings
from meld.sourceview import LanguageManager


@Gtk.Template(resource_path='/org/gnome/meld/ui/patch-dialog.ui')
class PatchDialog(Gtk.Dialog):

    __gtype_name__ = "PatchDialog"

    left_radiobutton = Gtk.Template.Child("left_radiobutton")
    reverse_checkbutton = Gtk.Template.Child("reverse_checkbutton")
    right_radiobutton = Gtk.Template.Child("right_radiobutton")
    side_selection_box = Gtk.Template.Child("side_selection_box")
    side_selection_label = Gtk.Template.Child("side_selection_label")
    textview: Gtk.TextView = Gtk.Template.Child("textview")

    def __init__(self, filediff):
        super().__init__()

        self.set_transient_for(filediff.get_toplevel())
        self.filediff = filediff

        buf = GtkSource.Buffer()
        self.textview.set_buffer(buf)
        lang = LanguageManager.get_language_from_mime_type("text/x-diff")
        buf.set_language(lang)
        buf.set_highlight_syntax(True)

        self.index_map = {self.left_radiobutton: (0, 1),
                          self.right_radiobutton: (1, 2)}
        self.left_patch = True
        self.reverse_patch = self.reverse_checkbutton.get_active()

        if self.filediff.num_panes < 3:
            self.side_selection_label.hide()
            self.side_selection_box.hide()

        meld_settings = get_meld_settings()
        self.textview.modify_font(meld_settings.font)
        self.textview.set_editable(False)
        meld_settings.connect('changed', self.on_setting_changed)

    def on_setting_changed(self, settings, key):
        if key == "font":
            self.textview.modify_font(settings.font)

    @Gtk.Template.Callback()
    def on_buffer_selection_changed(self, radiobutton):
        if not radiobutton.get_active():
            return
        self.left_patch = radiobutton == self.left_radiobutton
        self.update_patch()

    @Gtk.Template.Callback()
    def on_reverse_checkbutton_toggled(self, checkbutton):
        self.reverse_patch = checkbutton.get_active()
        self.update_patch()

    def update_patch(self):
        indices = (0, 1)
        if not self.left_patch:
            indices = (1, 2)
        if self.reverse_patch:
            indices = (indices[1], indices[0])

        texts = []
        for b in self.filediff.textbuffer:
            start, end = b.get_bounds()
            text = b.get_text(start, end, False)
            lines = text.splitlines(True)

            # Ensure that the last line ends in a newline
            barelines = text.splitlines(False)
            if barelines and lines and barelines[-1] == lines[-1]:
                # Final line lacks a line-break; add in a best guess
                if len(lines) > 1:
                    previous_linebreak = lines[-2][len(barelines[-2]):]
                else:
                    previous_linebreak = "\n"
                lines[-1] += previous_linebreak

            texts.append(lines)

        names = [self.filediff.textbuffer[i].data.label for i in range(3)]
        prefix = os.path.commonprefix(names)
        names = [n[prefix.rfind("/") + 1:] for n in names]

        buf = self.textview.get_buffer()
        text0, text1 = texts[indices[0]], texts[indices[1]]
        name0, name1 = names[indices[0]], names[indices[1]]

        diff = difflib.unified_diff(text0, text1, name0, name1)
        diff_text = "".join(d for d in diff)
        buf.set_text(diff_text)

    def save_patch(self, targetfile: Gio.File):
        buf = self.textview.get_buffer()
        sourcefile = GtkSource.File.new()
        saver = GtkSource.FileSaver.new_with_target(
            buf, sourcefile, targetfile)
        saver.save_async(
            GLib.PRIORITY_HIGH,
            callback=self.file_saved_cb,
        )

    def file_saved_cb(self, saver, result, *args):
        gfile = saver.get_location()
        try:
            saver.save_finish(result)
        except GLib.Error as err:
            filename = GLib.markup_escape_text(gfile.get_parse_name())
            error_dialog(
                primary=_("Could not save file %s.") % filename,
                secondary=_("Couldn’t save file due to:\n%s") % (
                    GLib.markup_escape_text(str(err))),
            )

    def run(self):
        self.update_patch()

        result = super().run()
        if result < 0:
            self.hide()
            return

        # Copy patch to clipboard
        if result == 1:
            buf = self.textview.get_buffer()
            start, end = buf.get_bounds()
            clip = Gtk.Clipboard.get_default(Gdk.Display.get_default())
            clip.set_text(buf.get_text(start, end, False), -1)
            clip.store()
        # Save patch as a file
        else:
            gfile = prompt_save_filename(_("Save Patch"))
            if gfile:
                self.save_patch(gfile)

        self.hide()



#! /usr/bin/python3

import glob
import os
import subprocess
import sys

from bs4 import BeautifulSoup

JEKYLL_HEADER = """---
layout: help
title: Meld - Help
---
"""

SCSS_HEADER = """
#help-content {
  border-left: solid 1px #e0e0df;
  border-right: solid 1px #e0e0df;
  background-color: #ffffff;
}

#help-content div.body {
  border: none !important; }

#help-content div.headbar {
  margin: 10px !important;
}

#help-content div.footbar {
  margin: 10px !important;
}

#help-content {

.title {
  line-height: 1em;
}

h1 {
  font-family: sans-serif;
  font-weight: bold;
  text-shadow: none;
  color: black;
}

h2 {
  font-family: sans-serif;
  text-shadow: none;
  color: black;
}
"""

SCSS_FOOTER = """
}
"""


def munge_html(filename):
    if not os.path.exists(filename):
        print("File not found: " + filename, file=sys.stderr)
        sys.exit(1)

    with open(filename) as f:
        contents = f.read()

    soup = BeautifulSoup(contents, "lxml")
    body = "".join([str(tag) for tag in soup.body])
    body = JEKYLL_HEADER + body

    print("Rewriting " + filename)
    with open(filename, "w") as f:
        f.write(body)


def munge_css(filename):

    if not os.path.exists(filename):
        print("File not found: " + filename, file=sys.stderr)
        sys.exit(1)

    with open(filename) as f:
        contents = f.read()

    contents = SCSS_HEADER + contents + SCSS_FOOTER
    new_css = sassify(contents)

    print("Rewriting " + filename)
    with open(filename, 'w') as f:
        f.write(new_css)


def sassify(scss_string):
    scss = subprocess.Popen(
        ['scss', '-s'],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        universal_newlines=True,
    )
    stdout, stderr = scss.communicate(scss_string)
    return stdout


if __name__ == "__main__":

    if os.path.exists('html'):
        print("Refusing to overwrite existing html/ folder", file=sys.stderr)
        sys.exit(1)

    print("Generating CSS with gnome-doc-tool...", file=sys.stderr)
    subprocess.check_call(['gnome-doc-tool', 'css'])

    print("Generating HTML with gnome-doc-tool...", file=sys.stderr)
    subprocess.check_call(['gnome-doc-tool', 'html', '-c', 'index.css',
                           '--copy-graphics', '*.page'])

    os.mkdir('html')
    for filename in glob.glob('*.html'):
        munge_html(filename)
        os.rename(filename, os.path.join('html', filename))

    munge_css('index.css')
    os.rename('index.css', os.path.join('html', 'index.css'))

    print("Embeddable documentation written to html/", file=sys.stderr)



# Copyright (C) 2002-2005 Stephen Kennedy <stevek@gnome.org>
# Copyright (C) 2015 Kai Willadsen <kai.willadsen@gmail.com>

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import errno
import os

from . import _vc


class Vc(_vc.Vc):

    CMD = "hg"
    NAME = "Mercurial"
    VC_DIR = ".hg"

    state_map = {
        "?": _vc.STATE_NONE,
        "A": _vc.STATE_NEW,
        "C": _vc.STATE_NORMAL,
        "!": _vc.STATE_MISSING,
        "I": _vc.STATE_IGNORED,
        "M": _vc.STATE_MODIFIED,
        "R": _vc.STATE_REMOVED,
    }

    def commit(self, runner, files, message):
        command = [self.CMD, 'commit', '-m', message]
        runner(command, files, refresh=True, working_dir=self.root)

    def update(self, runner):
        command = [self.CMD, 'pull', '-u']
        runner(command, [], refresh=True, working_dir=self.root)

    def add(self, runner, files):
        command = [self.CMD, 'add']
        runner(command, files, refresh=True, working_dir=self.root)

    def remove(self, runner, files):
        command = [self.CMD, 'rm']
        runner(command, files, refresh=True, working_dir=self.root)

    def revert(self, runner, files):
        command = [self.CMD, 'revert']
        runner(command, files, refresh=True, working_dir=self.root)

    @classmethod
    def valid_repo(cls, path):
        return not _vc.call([cls.CMD, "root"], cwd=path)

    def get_path_for_repo_file(self, path, commit=None):
        if commit is not None:
            raise NotImplementedError()

        if not path.startswith(self.root + os.path.sep):
            raise _vc.InvalidVCPath(self, path, "Path not in repository")
        path = path[len(self.root) + 1:]

        suffix = os.path.splitext(path)[1]
        args = [self.CMD, "cat", path]
        return _vc.call_temp_output(args, cwd=self.root, suffix=suffix)

    def _update_tree_state_cache(self, path):
        """ Update the state of the file(s) at self._tree_cache['path'] """
        while 1:
            try:
                # Get the status of modified files
                proc = _vc.popen([self.CMD, "status", '-A', path],
                                 cwd=self.location)
                entries = proc.read().split("\n")[:-1]

                # The following command removes duplicate file entries.
                # Just in case.
                entries = list(set(entries))
                break
            except OSError as e:
                if e.errno != errno.EAGAIN:
                    raise

        if len(entries) == 0 and os.path.isfile(path):
            # If we're just updating a single file there's a chance that it
            # was it was previously modified, and now has been edited so that
            # it is un-modified.  This will result in an empty 'entries' list,
            # and self._tree_cache['path'] will still contain stale data.
            # When this corner case occurs we force self._tree_cache['path']
            # to STATE_NORMAL.
            self._tree_cache[path] = _vc.STATE_NORMAL
        else:
            # There are 1 or more modified files, parse their state
            for entry in entries:
                # we might have a space in file name, it should be ignored
                statekey, name = entry.split(" ", 1)
                path = os.path.join(self.location, name.strip())
                state = self.state_map.get(statekey.strip(), _vc.STATE_NONE)
                self._tree_cache[path] = state
                self._add_missing_cache_entry(path, state)



from unittest import mock

import pytest
from gi.repository import Gio

from meld.iohelpers import (
    find_shared_parent_path,
    format_home_relative_path,
    format_parent_relative_path,
)


@pytest.mark.parametrize(
    'paths, expected_parent',
    [
        # No paths, None return
        ([], None),
        # One path always returns its own parent
        (['/foo/a/b/c'], '/foo/a/b'),
        # Two paths
        (['/foo/a', '/foo/b'], '/foo'),
        # Three paths
        (['/foo/a', '/foo/b', '/foo/c'], '/foo'),
        # First path is deeper
        (['/foo/a/asd/asd', '/foo/b'], '/foo'),
        # Second path is deeper
        (['/foo/a/', '/foo/b/asd/asd'], '/foo'),
        # Common parent is the root
        (['/foo/a/', '/bar/b/'], '/'),
        # One path, one missing path
        (['/foo/a', None], None),
        # Two paths, one missing path
        (['/foo/a', None, '/foo/c'], None),
    ],
)
def test_find_shared_parent_path(paths, expected_parent):
    files = [Gio.File.new_for_path(p) if p else None for p in paths]
    print([f.get_path() if f else repr(f) for f in files])
    parent = find_shared_parent_path(files)

    if parent is None:
        assert expected_parent is None
    else:
        print(f'Parent: {parent.get_path()}; expected {expected_parent}')
        if expected_parent is None:
            assert parent is None
        else:
            assert parent.equal(Gio.File.new_for_path(expected_parent))


@pytest.mark.parametrize(
    "path, expected_format",
    [
        ("/home/hey/foo", "~/foo"),
        ("/home/hmph/foo", "/home/hmph/foo"),
    ]
)
def test_format_home_relative_path(path, expected_format):

    with mock.patch(
        "meld.iohelpers.GLib.get_home_dir",
        return_value="/home/hey/",
    ):
        gfile = Gio.File.new_for_path(path)
        assert format_home_relative_path(gfile) == expected_format


@pytest.mark.parametrize(
    'parent, child, expected_label',
    [
        # Child is a direct child of parent
        (
            '/home/hey/',
            '/home/hey/foo.txt',
            '…/hey/foo.txt',
        ),
        # Child is a direct child of parent and parent is the root
        (
            '/',
            '/foo.txt',
            '/foo.txt',
        ),
        # Child is a 2-depth child of parent
        (
            '/home/hey/',
            '/home/hey/project/foo.txt',
            '…/project/foo.txt',
        ),
        # Child is a 3-depth child of parent
        (
            '/home/hey/',
            '/home/hey/project/package/foo.txt',
            '…/project/package/foo.txt',
        ),
        # Child is a more-than-3-depth child of parent, with long
        # immediate parent
        (
            '/home/hey/',
            '/home/hey/project/package/subpackage/foo.txt',
            '…/project/…/subpackage/foo.txt',
        ),
        # Child is a more-than-3-depth child of parent, with short
        # immediate parent
        (
            '/home/hey/',
            '/home/hey/project/package/src/foo.txt',
            '…/project/package/src/foo.txt',
        ),
    ],
)
def test_format_parent_relative_path(
    parent: str,
    child: str,
    expected_label: str,
):
    parent_gfile = Gio.File.new_for_path(parent)
    child_gfile = Gio.File.new_for_path(child)

    label = format_parent_relative_path(parent_gfile, child_gfile)

    assert label == expected_label


def test_format_parent_relative_path_no_parent():
    parent_gfile = Gio.File.new_for_path('/')
    child_gfile = Gio.File.new_for_path('/')

    with pytest.raises(ValueError, match='has no parent'):
        format_parent_relative_path(parent_gfile, child_gfile)
